@@ -1,4 +1,3 @@
-#!/usr/bin/env python
 # -*- Mode: python; tab-width: 4; indent-tabs-mode: nil -*-
 #
 # This program is free software; you can redistribute it and/or modify it under
@@ -15,7 +14,7 @@
 # with this program; if not, write to the Free Software Foundation, Inc., 51
 # Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 #
-# Copyright (C) 2013 - 2014 Aleksander Morgado <aleksander@aleksander.es>
+# Copyright (C) 2013 - 2018 Aleksander Morgado <aleksander@aleksander.es>
 #
 
 import string
@@ -73,7 +72,7 @@ class Struct:
                     ' * @${field_name_underscore}: a #MbimUuid.\n')
             elif field['format'] =='byte-array':
                 inner_template = (' * @${field_name_underscore}: an array of #guint8 values.\n')
-            elif field['format'] =='unsized-byte-array' or field['format'] == 'ref-byte-array':
+            elif field['format'] in ['unsized-byte-array', 'ref-byte-array', 'ref-byte-array-no-offset']:
                 inner_template = ''
                 if 'array-size-field' not in field:
                     inner_template += (' * @${field_name_underscore}_size: size of the ${field_name_underscore} array.\n')
@@ -99,18 +98,12 @@ class Struct:
             elif field['format'] == 'ref-ipv4':
                 inner_template = (
                     ' * @${field_name_underscore}: a #MbimIPv4.\n')
-            elif field['format'] == 'ipv4-array':
-                inner_template = (
-                    ' * @${field_name_underscore}: an array of #MbimIPv4 values.\n')
             elif field['format'] == 'ipv6':
                 inner_template = (
                     ' * @${field_name_underscore}: a #MbimIPv6\n')
             elif field['format'] == 'ref-ipv6':
                 inner_template = (
                     ' * @${field_name_underscore}: a #MbimIPv6\n')
-            elif field['format'] == 'ipv6-array':
-                inner_template = (
-                    ' * @${field_name_underscore}: an array of #MbimIPv6 values.\n')
             else:
                 raise ValueError('Cannot handle format \'%s\' in struct' % field['format'])
             template += string.Template(inner_template).substitute(translations)
@@ -127,7 +120,7 @@ class Struct:
                 translations['array_size'] = field['array-size']
                 inner_template = (
                     '    guint8 ${field_name_underscore}[${array_size}];\n')
-            elif field['format'] == 'unsized-byte-array' or field['format'] == 'ref-byte-array':
+            elif field['format'] in ['unsized-byte-array', 'ref-byte-array', 'ref-byte-array-no-offset']:
                 inner_template = ''
                 if 'array-size-field' not in field:
                     inner_template += (
@@ -155,18 +148,12 @@ class Struct:
             elif field['format'] == 'ref-ipv4':
                 inner_template = (
                     '    MbimIPv4 ${field_name_underscore};\n')
-            elif field['format'] == 'ipv4-array':
-                inner_template = (
-                    '    MbimIPv4 *${field_name_underscore};\n')
             elif field['format'] == 'ipv6':
                 inner_template = (
                     '    MbimIPv6 ${field_name_underscore};\n')
             elif field['format'] == 'ref-ipv6':
                 inner_template = (
                     '    MbimIPv6 ${field_name_underscore};\n')
-            elif field['format'] == 'ipv6-array':
-                inner_template = (
-                    '    MbimIPv6 *${field_name_underscore};\n')
             else:
                 raise ValueError('Cannot handle format \'%s\' in struct' % field['format'])
             template += string.Template(inner_template).substitute(translations)
@@ -204,7 +191,7 @@ class Struct:
             inner_template = ''
             if field['format'] == 'uuid':
                 pass
-            elif field['format'] == 'unsized-byte-array' or field['format'] == 'ref-byte-array':
+            elif field['format'] in ['unsized-byte-array', 'ref-byte-array', 'ref-byte-array-no-offset']:
                 inner_template += (
                     '    g_free (var->${field_name_underscore});\n')
             elif field['format'] == 'guint32':
@@ -224,16 +211,10 @@ class Struct:
                 pass
             elif field['format'] == 'ref-ipv4':
                 pass
-            elif field['format'] == 'ipv4-array':
-                inner_template += (
-                    '    g_free (var->${field_name_underscore});\n')
             elif field['format'] == 'ipv6':
                 pass
             elif field['format'] == 'ref-ipv6':
                 pass
-            elif field['format'] == 'ipv6-array':
-                inner_template += (
-                    '    g_free (var->${field_name_underscore});\n')
             else:
                 raise ValueError('Cannot handle format \'%s\' in struct clear' % field['format'])
             template += string.Template(inner_template).substitute(translations)
@@ -287,6 +268,134 @@ class Struct:
                 '}\n')
             cfile.write(string.Template(template).substitute(translations))
 
+    """
+    Emit the type's print methods
+    """
+    def _emit_print(self, cfile):
+        translations = { 'name'            : self.name,
+                         'name_underscore' : utils.build_underscore_name_from_camelcase(self.name),
+                         'struct_size'     : self.size }
+
+        template = (
+            '\n'
+            'static gchar *\n'
+            '_mbim_message_print_${name_underscore}_struct (\n'
+            '    const ${name} *self,\n'
+            '    const gchar *line_prefix)\n'
+            '{\n'
+            '    GString *str;\n'
+            '\n'
+            '    str = g_string_new ("");\n'
+            '\n')
+
+        for field in self.contents:
+            translations['field_name'] = field['name']
+            translations['field_name_underscore'] = utils.build_underscore_name_from_camelcase(field['name'])
+
+            inner_template = (
+                '    g_string_append_printf (str, "%s  ${field_name} = ", line_prefix);\n'
+                '    {\n')
+
+            if field['format'] == 'uuid':
+                inner_template += (
+                    '        gchar *tmpstr;\n'
+                    '\n'
+                    '        tmpstr = mbim_uuid_get_printable (&(self->${field_name_underscore}));\n'
+                    '        g_string_append_printf (str, "\'%s\'", tmpstr);\n'
+                    '        g_free (tmpstr);\n')
+
+            elif field['format'] in ['byte-array', 'ref-byte-array', 'ref-byte-array-no-offset', 'unsized-byte-array']:
+                inner_template += (
+                    '        guint i;\n'
+                    '        guint array_size;\n'
+                    '\n')
+
+                if field['format'] == 'byte-array':
+                    translations['array_size'] = field['array-size']
+                    inner_template += (
+                        '        array_size = ${array_size};\n')
+                elif 'array-size-field' in field:
+                    translations['array_size_field_name_underscore'] = utils.build_underscore_name_from_camelcase(field['array-size-field'])
+                    inner_template += (
+                        '        array_size = self->${array_size_field_name_underscore};\n')
+                else:
+                    inner_template += (
+                        '        array_size = self->${field_name_underscore}_size;\n')
+
+                inner_template += (
+                    '        g_string_append (str, "\'");\n'
+                    '        for (i = 0; i < array_size; i++)\n'
+                    '            g_string_append_printf (str, "%02x%s", self->${field_name_underscore}[i], (i == (array_size - 1)) ? "" : ":" );\n'
+                    '        g_string_append (str, "\'");\n')
+
+            elif field['format'] == 'guint32':
+                inner_template += (
+                    '        g_string_append_printf (str, "\'%" G_GUINT32_FORMAT "\'", self->${field_name_underscore});\n')
+
+            elif field['format'] == 'guint64':
+                inner_template += (
+                    '        g_string_append_printf (str, "\'%" G_GUINT64_FORMAT "\'", self->${field_name_underscore});\n')
+
+            elif field['format'] == 'guint32-array':
+                translations['array_size_field_name_underscore'] = utils.build_underscore_name_from_camelcase(field['array-size-field'])
+                inner_template += (
+                    '        guint i;\n'
+                    '\n'
+                    '        g_string_append (str, "\'");\n'
+                    '        for (i = 0; i < self->${array_size_field_name_underscore}; i++)\n'
+                    '            g_string_append_printf (str, "%" G_GUINT32_FORMAT "%s", self->${field_name_underscore}[i], (i == (self->${array_size_field_name_underscore} - 1)) ? "" : "," );\n'
+                    '        g_string_append (str, "\'");\n')
+
+            elif field['format'] == 'string':
+                inner_template += (
+                    '        g_string_append_printf (str, "\'%s\'", self->${field_name_underscore});\n')
+
+            elif field['format'] == 'string-array':
+                translations['array_size_field_name_underscore'] = utils.build_underscore_name_from_camelcase(field['array-size-field'])
+                inner_template += (
+                    '        guint i;\n'
+                    '\n'
+                    '        g_string_append (str, "\'");\n'
+                    '        for (i = 0; i < self->${array_size_field_name_underscore}; i++)\n'
+                    '            g_string_append_printf (str, "%s%s", self->${field_name_underscore}[i], (i == (self->${array_size_field_name_underscore} - 1)) ? "" : "," );\n'
+                    '        g_string_append (str, "\'");\n')
+
+            elif field['format'] == 'ipv4' or \
+                 field['format'] == 'ref-ipv4' or \
+                 field['format'] == 'ipv6' or \
+                 field['format'] == 'ref-ipv6':
+                inner_template += (
+                    '        GInetAddress *addr;\n'
+                    '        gchar *tmpstr;\n'
+                    '\n')
+
+                if field['format'] == 'ipv4' or \
+                   field['format'] == 'ref-ipv4':
+                    inner_template += (
+                        '        addr = g_inet_address_new_from_bytes ((guint8 *)&(self->${field_name_underscore}.addr), G_SOCKET_FAMILY_IPV4);\n')
+                elif field['format'] == 'ipv6' or \
+                   field['format'] == 'ref-ipv6':
+                    inner_template += (
+                        '        addr = g_inet_address_new_from_bytes ((guint8 *)&(self->${field_name_underscore}.addr), G_SOCKET_FAMILY_IPV6);\n')
+
+                inner_template += (
+                    '        tmpstr = g_inet_address_to_string (addr);\n'
+                    '        g_string_append_printf (str, "\'%s\'", tmpstr);\n'
+                    '        g_free (tmpstr);\n'
+                    '        g_object_unref (addr);\n')
+
+            else:
+                raise ValueError('Cannot handle format \'%s\' in struct' % field['format'])
+
+            inner_template += (
+                '    }\n'
+                '    g_string_append (str, "\\n");\n')
+            template += (string.Template(inner_template).substitute(translations))
+
+        template += (
+            '    return g_string_free (str, FALSE);\n'
+            '}\n')
+        cfile.write(string.Template(template).substitute(translations))
 
     """
     Emit the type's read methods
@@ -320,7 +429,8 @@ class Struct:
                     '\n'
                     '    memcpy (&(out->${field_name_underscore}), _mbim_message_read_uuid (self, offset), 16);\n'
                     '    offset += 16;\n')
-            elif field['format'] == 'ref-byte-array':
+            elif field['format'] in ['ref-byte-array', 'ref-byte-array-no-offset']:
+                translations['has_offset'] = 'TRUE' if field['format'] == 'ref-byte-array' else 'FALSE'
                 if 'array-size-field' in field:
                     translations['array_size_field_name_underscore'] = utils.build_underscore_name_from_camelcase(field['array-size-field'])
                     inner_template += (
@@ -328,7 +438,7 @@ class Struct:
                         '    {\n'
                         '        const guint8 *tmp;\n'
                         '\n'
-                        '        tmp = _mbim_message_read_byte_array (self, relative_offset, offset, TRUE, FALSE, NULL);\n'
+                        '        tmp = _mbim_message_read_byte_array (self, relative_offset, offset, ${has_offset}, FALSE, NULL);\n'
                         '        out->${field_name_underscore} = g_malloc (out->${array_size_field_name_underscore});\n'
                         '        memcpy (out->${field_name_underscore}, tmp, out->${array_size_field_name_underscore});\n'
                         '        offset += 4;\n'
@@ -339,7 +449,7 @@ class Struct:
                         '    {\n'
                         '        const guint8 *tmp;\n'
                         '\n'
-                        '        tmp = _mbim_message_read_byte_array (self, relative_offset, offset, TRUE, TRUE, &(out->${field_name_underscore}_size));\n'
+                        '        tmp = _mbim_message_read_byte_array (self, relative_offset, offset, ${has_offset}, TRUE, &(out->${field_name_underscore}_size));\n'
                         '        out->${field_name_underscore} = g_malloc (out->${field_name_underscore}_size);\n'
                         '        memcpy (out->${field_name_underscore}, tmp, out->${field_name_underscore}_size);\n'
                         '        offset += 8;\n'
@@ -403,26 +513,16 @@ class Struct:
                     '\n'
                     '    memcpy (&(out->${field_name_underscore}), _mbim_message_read_ipv4 (self, offset, TRUE), 4);\n'
                     '    offset += 4;\n')
-            elif field['format'] == 'ipv4-array':
-                inner_template += (
-                    '\n'
-                    '    out->${field_name_underscore} =_mbim_message_read_ipv4_array (self, out->${array_size_field_name_underscore}, offset);\n'
-                    '    offset += 4;\n')
             elif field['format'] == 'ipv6':
                 inner_template += (
                     '\n'
                     '    memcpy (&(out->${field_name_underscore}), _mbim_message_read_ipv6 (self, offset, FALSE), 16);\n'
-                    '    offset += 4;\n')
+                    '    offset += 16;\n')
             elif field['format'] == 'ref-ipv6':
                 inner_template += (
                     '\n'
                     '    memcpy (&(out->${field_name_underscore}), _mbim_message_read_ipv6 (self, offset, TRUE), 16);\n'
                     '    offset += 4;\n')
-            elif field['format'] == 'ipv6-array':
-                inner_template += (
-                    '\n'
-                    '    out->${field_name_underscore} =_mbim_message_read_ipv6_array (self, out->${array_size_field_name_underscore}, offset);\n'
-                    '    offset += 4;\n')
             else:
                 raise ValueError('Cannot handle format \'%s\' in struct' % field['format'])
 
@@ -502,19 +602,18 @@ class Struct:
                 inner_template = ('    _mbim_struct_builder_append_byte_array (builder, FALSE, FALSE, ${pad_array}, value->${field}, ${array_size});\n')
             elif field['format'] == 'unsized-byte-array':
                 inner_template = ('    _mbim_struct_builder_append_byte_array (builder, FALSE, FALSE, ${pad_array}, value->${field}, value->${field}_size);\n')
-            elif field['format'] == 'ref-byte-array':
+            elif field['format'] in ['ref-byte-array', 'ref-byte-array-no-offset']:
+                translations['has_offset'] = 'TRUE' if field['format'] == 'ref-byte-array' else 'FALSE'
                 if 'array-size-field' in field:
-                    inner_template = ('    _mbim_struct_builder_append_byte_array (builder, TRUE, FALSE, ${pad_array}, value->${field}, value->${array_size_field});\n')
+                    inner_template = ('    _mbim_struct_builder_append_byte_array (builder, ${has_offset}, FALSE, ${pad_array}, value->${field}, value->${array_size_field});\n')
                 else:
-                    inner_template = ('    _mbim_struct_builder_append_byte_array (builder, TRUE, TRUE, ${pad_array}, value->${field}, value->${field}_size);\n')
+                    inner_template = ('    _mbim_struct_builder_append_byte_array (builder, ${has_offset}, TRUE, ${pad_array}, value->${field}, value->${field}_size);\n')
             elif field['format'] == 'guint32':
                 inner_template = ('    _mbim_struct_builder_append_guint32 (builder, value->${field});\n')
             elif field['format'] == 'guint32-array':
                 inner_template = ('    _mbim_struct_builder_append_guint32_array (builder, value->${field}, value->${array_size_field});\n')
             elif field['format'] == 'guint64':
                 inner_template = ('    _mbim_struct_builder_append_guint64 (builder, value->${field});\n')
-            elif field['format'] == 'guint64-array':
-                inner_template = ('    _mbim_struct_builder_append_guint64_array (builder, value->${field}, value->${array_size_field});\n')
             elif field['format'] == 'string':
                 inner_template = ('    _mbim_struct_builder_append_string (builder, value->${field});\n')
             elif field['format'] == 'string-array':
@@ -523,14 +622,10 @@ class Struct:
                 inner_template = ('    _mbim_struct_builder_append_ipv4 (builder, &value->${field}, FALSE);\n')
             elif field['format'] == 'ref-ipv4':
                 inner_template = ('    _mbim_struct_builder_append_ipv4 (builder, &value->${field}, TRUE);\n')
-            elif field['format'] == 'ipv4-array':
-                inner_template = ('    _mbim_struct_builder_append_ipv4_array (builder, value->${field}, value->${array_size_field});\n')
             elif field['format'] == 'ipv6':
                 inner_template = ('    _mbim_struct_builder_append_ipv6 (builder, &value->${field}, FALSE);\n')
             elif field['format'] == 'ref-ipv6':
                 inner_template = ('    _mbim_struct_builder_append_ipv6 (builder, &value->${field}, TRUE);\n')
-            elif field['format'] == 'ipv6-array':
-                inner_template = ('    _mbim_struct_builder_append_ipv6_array (builder, value->${field}, value->${array_size_field});\n')
             else:
                 raise ValueError('Cannot handle format \'%s\' in struct' % field['format'])
 
@@ -667,6 +762,8 @@ class Struct:
         self._emit_free(hfile, cfile)
         # Emit type's read
         self._emit_read(cfile)
+        # Emit type's print
+        self._emit_print(cfile)
         # Emit type's append
         self._emit_append(cfile)
 
